@@ -3,7 +3,7 @@
   Copyright (C) 2020, 2021 Anastasis SARL
 
   Anastasis is free software; you can redistribute it and/or modify it under the
-  terms of the GNU Lesser General Public License as published by the Free Software
+  terms of the GNU Affero General Public License as published by the Free Software
   Foundation; either version 3, or (at your option) any later version.
 
   Anastasis is distributed in the hope that it will be useful, but WITHOUT ANY
@@ -106,8 +106,6 @@ AH_handler_config (struct AH_RequestHandler *rh,
                              "0:0:0"),
     GNUNET_JSON_pack_string ("business_name",
                              AH_business_name),
-    GNUNET_JSON_pack_string ("currency",
-                             (char *) AH_currency),
     GNUNET_JSON_pack_array_steal ("methods",
                                   method_arr),
     GNUNET_JSON_pack_uint64 ("storage_limit_in_megabytes",
@@ -118,8 +116,8 @@ AH_handler_config (struct AH_RequestHandler *rh,
                             &AH_truth_upload_fee),
     TALER_JSON_pack_amount ("liability_limit",
                             &AH_insurance),
-    GNUNET_JSON_pack_data_auto ("server_salt",
-                                &AH_server_salt));
+    GNUNET_JSON_pack_data_auto ("provider_salt",
+                                &AH_provider_salt));
 }
 
 
