@@ -169,11 +169,19 @@ handle_truth_store_finished (void *cls,
     ud.ec = TALER_JSON_get_error_code2 (data,
                                         data_size);
     break;
-  default:
-    GNUNET_break (0);
+  case MHD_HTTP_BAD_GATEWAY:
     ud.ec = TALER_JSON_get_error_code2 (data,
                                         data_size);
     break;
+  default:
+    ud.ec = TALER_JSON_get_error_code2 (data,
+                                        data_size);
+    GNUNET_break (0);
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "Unexpected HTTP status code %u/%d\n",
+                (unsigned int) response_code,
+                ud.ec);
+    break;
   }
   tso->cb (tso->cb_cls,
            &ud);
