@@ -1,6 +1,6 @@
 --
 -- This file is part of Anastasis
--- Copyright (C) 2020, 2021 Anastasis SARL SA
+-- Copyright (C) 2020, 2021, 2022 Anastasis SARL SA
 --
 -- ANASTASIS is free software; you can redistribute it and/or modify it under the
 -- terms of the GNU General Public License as published by the Free Software
@@ -20,6 +20,11 @@ BEGIN;
 -- Check patch versioning is in place.
 SELECT _v.register_patch('stasis-0001', NULL, NULL);
 
+CREATE SCHEMA anastasis;
+COMMENT ON SCHEMA anastasis IS 'anastasis backend data';
+
+SET search_path TO anastasis;
+
 
 CREATE TABLE IF NOT EXISTS anastasis_truth_payment
   (truth_uuid BYTEA PRIMARY KEY CHECK(LENGTH(truth_uuid)=32),
@@ -40,7 +45,7 @@ COMMENT ON COLUMN anastasis_truth_payment.expiration
 
 CREATE TABLE IF NOT EXISTS anastasis_truth
   (truth_uuid BYTEA PRIMARY KEY CHECK(LENGTH(truth_uuid)=32),
-   key_share_data BYTEA CHECK(LENGTH(key_share_data)=80) NOT NULL,
+   key_share_data BYTEA CHECK(LENGTH(key_share_data)=72) NOT NULL,
    method_name VARCHAR NOT NULL,
    encrypted_truth BYTEA NOT NULL,
    truth_mime VARCHAR NOT NULL,
@@ -73,7 +78,7 @@ COMMENT ON COLUMN anastasis_user.expiration_date
 
 
 CREATE TABLE IF NOT EXISTS anastasis_recdoc_payment
-  (payment_id BIGSERIAL PRIMARY KEY,
+  (payment_id BIGINT GENERATED BY DEFAULT AS IDENTITY PRIMARY KEY,
    user_id BYTEA NOT NULL REFERENCES anastasis_user(user_id),
    post_counter INT4 NOT NULL DEFAULT 0 CHECK(post_counter >= 0),
    amount_val INT8 NOT NULL,
@@ -102,7 +107,7 @@ COMMENT ON COLUMN anastasis_recdoc_payment.paid
 
 
 CREATE TABLE IF NOT EXISTS anastasis_challenge_payment
-  (payment_id BIGSERIAL PRIMARY KEY,
+  (payment_id BIGINT GENERATED BY DEFAULT AS IDENTITY PRIMARY KEY,
    truth_uuid BYTEA CHECK(LENGTH(truth_uuid)=32) NOT NULL,
    amount_val INT8 NOT NULL,
    amount_frac INT4 NOT NULL,
@@ -140,6 +145,8 @@ CREATE TABLE IF NOT EXISTS anastasis_recoverydocument
    account_sig BYTEA NOT NULL CHECK(LENGTH(account_sig)=64),
    recovery_data_hash BYTEA NOT NULL CHECK(length(recovery_data_hash)=64),
    recovery_data BYTEA NOT NULL,
+   recovery_meta_data BYTEA NOT NULL,
+   creation_date INT8 NOT NULL,
    PRIMARY KEY (user_id, version));
 COMMENT ON TABLE anastasis_recoverydocument
   IS 'Stores a recovery document which contains the policy and the encrypted core secret';
@@ -151,8 +158,12 @@ COMMENT ON COLUMN anastasis_recoverydocument.account_sig
   IS 'Signature of the recovery document';
 COMMENT ON COLUMN anastasis_recoverydocument.recovery_data_hash
   IS 'Hash of the recovery document to prevent unnecessary uploads';
+COMMENT ON COLUMN anastasis_recoverydocument.creation_date
+  IS 'Creation date of the recovery document (when it was uploaded)';
 COMMENT ON COLUMN anastasis_recoverydocument.recovery_data
   IS 'Contains the encrypted policy and core secret';
+COMMENT ON COLUMN anastasis_recoverydocument.recovery_meta_data
+  IS 'Contains an encrypted human-readable and sometimes user-generated description of the backup';
 
 
 CREATE TABLE IF NOT EXISTS anastasis_challengecode
@@ -194,7 +205,7 @@ COMMENT ON INDEX anastasis_challengecode_expiration_index
 
 
 CREATE TABLE IF NOT EXISTS anastasis_auth_iban_in
-  (auth_in_serial_id BIGSERIAL UNIQUE
+  (auth_in_serial_id BIGINT GENERATED BY DEFAULT AS IDENTITY UNIQUE
   ,wire_reference INT8 NOT NULL PRIMARY KEY
   ,wire_subject TEXT NOT NULL
   ,credit_val INT8 NOT NULL
