@@ -3,7 +3,7 @@
   Copyright (C) 2020, 2021 Anastasis SARL
 
   Anastasis is free software; you can redistribute it and/or modify it under the
-  terms of the GNU Lesser General Public License as published by the Free Software
+  terms of the GNU Affero General Public License as published by the Free Software
   Foundation; either version 3, or (at your option) any later version.
 
   Anastasis is distributed in the hope that it will be useful, but WITHOUT ANY
@@ -151,7 +151,7 @@ run (void *cls)
                                          &paymentSecretP,
                                          &amount));
   {
-    struct GNUNET_TIME_Absolute res_time;
+    struct GNUNET_TIME_Timestamp res_time;
 
     FAILIF (GNUNET_DB_STATUS_SUCCESS_ONE_RESULT !=
             plugin->increment_lifetime (plugin->cls,
@@ -192,11 +192,13 @@ run (void *cls)
                                            &recoveryDataHash,
                                            recovery_data,
                                            strlen (recovery_data),
+                                           "meta-data",
+                                           strlen ("meta-data"),
                                            &paymentSecretP,
                                            &docVersion));
   {
     uint32_t vrs;
-    struct GNUNET_TIME_Absolute exp;
+    struct GNUNET_TIME_Timestamp exp;
 
     FAILIF (ANASTASIS_DB_ACCOUNT_STATUS_VALID_HASH_RETURNED !=
             plugin->lookup_account (plugin->cls,
@@ -240,7 +242,7 @@ run (void *cls)
   GNUNET_free (res_recovery_data);
 
   {
-    struct GNUNET_TIME_Absolute rt;
+    struct GNUNET_TIME_Timestamp rt;
 
     FAILIF (GNUNET_DB_STATUS_SUCCESS_ONE_RESULT !=
             plugin->create_challenge_code (plugin->cls,
@@ -250,10 +252,10 @@ run (void *cls)
                                            3, /* retry counter */
                                            &rt,
                                            &challenge_code));
-    FAILIF (0 != rt.abs_value_us);
+    FAILIF (! GNUNET_TIME_absolute_is_zero (rt.abs_time));
   }
   {
-    struct GNUNET_TIME_Absolute rt;
+    struct GNUNET_TIME_Timestamp rt;
     uint64_t c2;
 
     FAILIF (GNUNET_DB_STATUS_SUCCESS_ONE_RESULT !=
